from shexer.core.instances.annotators.strategy_mode.base_strategy_mode import BaseStrategyMode


class CompoundStrategyMode(BaseStrategyMode):

    def __init__(self, annotator_ref, list_of_strategies):
        super().__init__(annotator_ref)
        self._list_of_strategies = list_of_strategies


    def is_relevant_triple(self, a_triple):
        for a_strategy in self._list_of_strategies:
            if a_strategy.is_relevant_triple(a_triple):
                return True
        return False

    def annotate_triple(self, a_triple):
        for a_strategy in self._list_of_strategies:
            a_strategy.annotate_triple(a_triple)

    def annotation_post_parsing(self):
        for a_strategy in self._list_of_strategies:
            a_strategy.annotation_post_parsing()



from shexer.model.statement import POSITIVE_CLOSURE, KLEENE_CLOSURE, OPT_CARDINALITY
from shexer.model.IRI import IRI_ELEM_TYPE
from shexer.model.fixed_prop_choice_statement import FixedPropChoiceStatement
from shexer.io.shex.formater.statement_serializers.st_serializers_factory import StSerializerFactory
from shexer.core.shexing.strategy.minimal_iri_strategy.annotate_min_iri_strategy import AnnotateMinIriStrategy
from shexer.core.shexing.strategy.minimal_iri_strategy.ignore_min_iri_strategy import IgnoreMinIriStrategy


_DIRECT_ST_SERIALIZER = 0
_INVERSE_ST_SERIALIZER = 1


class AbstractShexingStrategy(object):

    def __init__(self, class_shexer):
        self._class_shexer = class_shexer
        self._namespaces_dict = class_shexer._namespaces_dict
        self._allow_opt_cardinality = class_shexer._allow_opt_cardinality
        self._disable_comments = self._class_shexer._disable_comments
        self._instantiation_property_str = self._class_shexer._instantiation_property_str
        self._keep_less_specific = self._class_shexer._keep_less_specific
        self._discard_useless_positive_closures = self._class_shexer._discard_useless_positive_closures
        self._tolerance = self._class_shexer._tolerance
        self._disable_or_statements = self._class_shexer._disable_or_statements
        self._all_compliant_mode = self._class_shexer._all_compliant_mode
        self._disable_exact_cardinality = self._class_shexer._disable_exact_cardinality
        self._allow_redundant_or = self._class_shexer._allow_redundant_or

        self._strategy_min_iri = AnnotateMinIriStrategy(class_shexer._class_min_iris_dict) \
            if class_shexer._detect_minimal_iri \
            else IgnoreMinIriStrategy()

        self._statement_serializer_factory = StSerializerFactory(freq_mode=class_shexer._instances_report_mode,
                                                                 decimals=class_shexer._decimals,
                                                                 instantiation_property_str=self._instantiation_property_str,
                                                                 disable_comments=self._disable_comments)


    def yield_base_shapes(self, acceptance_threshold):
        for a_shape in self._yield_base_shapes_direction_aware(acceptance_threshold=acceptance_threshold):
            self._strategy_min_iri.annotate_shape_iri(a_shape)
            yield a_shape

    def _yield_base_shapes_direction_aware(self, acceptance_threshold):
        raise NotImplementedError()

    def set_valid_shape_constraints(self, shape):
        raise NotImplementedError()

    def remove_statements_to_gone_shapes(self, shape, shape_names_to_remove):
        raise NotImplementedError()

    def _tune_list_of_valid_statements(self, valid_statements):
        """
        This method modifies the statements objects received --> no return needed
        :param valid_statements:
        :return:
        """
        if not len(valid_statements) == 0:
            valid_statements.sort(reverse=True, key=lambda x: x.probability)  # Restoring order completely
                                                                              # before changing cardinalities

            if self._all_compliant_mode:
                self._modify_cardinalities_of_statements_non_compliant_with_all_instances(valid_statements)

            if self._disable_exact_cardinality:
                self._generalize_exact_cardinalities(valid_statements)

            if self._disable_comments:
                self._remove_comments_from_statements(valid_statements)

    def _select_valid_statements_of_shape(self, original_statements):
        if len(original_statements) == 0:
            return []

        for a_statement in original_statements:  # TODO Refactor!!! This is not the place to set the serializer
            self._set_serializer_object_for_statements(a_statement)

        result = self._group_constraints_with_same_prop_and_obj(original_statements)
        result = self._group_IRI_constraints(result)

        return result

    def _compute_frequency(self, number_of_instances, n_ocurrences_statement):
        return float(n_ocurrences_statement) / number_of_instances

    def _modify_cardinalities_of_statements_non_compliant_with_all_instances(self, statements):
        for a_statement in statements:
            if a_statement.probability != 1:
                self._change_statement_cardinality_to_all_compliant(a_statement)

    def _change_statement_cardinality_to_all_compliant(self, statement):
        comment_for_current_sentence = self._turn_statement_into_comment(statement)
        statement.add_comment(comment=comment_for_current_sentence,
                              insert_first=True)
        statement.cardinality = OPT_CARDINALITY if \
            self._allow_opt_cardinality and statement.cardinality == 1 \
            else KLEENE_CLOSURE
        statement.probability = 1

    def _turn_statement_into_comment(self, a_statement):
        return a_statement.comment_representation(namespaces_dict=self._namespaces_dict)

    def _generalize_exact_cardinalities(self, statements):
        for a_statement in statements:
            if type(a_statement.cardinality) == int and a_statement.cardinality > 1:
                a_statement.cardinality = POSITIVE_CLOSURE

    def _remove_comments_from_statements(self, valid_statements):
        for a_statement in valid_statements:
            a_statement.remove_comments()

    def _set_serializer_object_for_statements(self, statement):
        statement.serializer_object = self._statement_serializer_factory.get_base_serializer(
            is_inverse=statement.is_inverse
        )

    def _group_constraints_with_same_prop_and_obj(self, candidate_statements):
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement not in already_visited:
                already_visited.add(a_statement)
                group_to_decide = [a_statement]
                for j in range(i + 1, len(candidate_statements)):
                    if self._statements_have_same_tokens(a_statement,
                                                         candidate_statements[j]):
                        group_to_decide.append(candidate_statements[j])
                        already_visited.add(candidate_statements[j])
                if len(group_to_decide) == 1:
                    result.append(a_statement)
                else:
                    result.append(self._decide_best_statement_with_cardinalities_in_comments(group_to_decide))
        return result

    def _statements_have_same_tokens(self, st1, st2):
        if st1.st_property == st2.st_property and st1.st_type == st2.st_type:
            return True
        return False

    def _decide_best_statement_with_cardinalities_in_comments(self, list_of_candidate_statements):
        if self._discard_useless_positive_closures:
            if self._is_a_group_of_statements_with_useless_positive_closure(list_of_candidate_statements):
                return self._statement_for_a_group_with_a_useless_positive_closure(list_of_candidate_statements)
        list_of_candidate_statements.sort(reverse=True, key=lambda x: x.probability)
        result = None
        if self._keep_less_specific:
            for a_statement in list_of_candidate_statements:
                if a_statement.cardinality == POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = list_of_candidate_statements[0]
        else:
            for a_statement in list_of_candidate_statements:
                if a_statement.cardinality != POSITIVE_CLOSURE:
                    result = a_statement
                    break
            if result is None:
                result = list_of_candidate_statements[0]

        for a_statement in list_of_candidate_statements:
            if a_statement.cardinality != result.cardinality:
                result.add_comment(self._turn_statement_into_comment(a_statement))
        return result

    def _is_a_group_of_statements_with_useless_positive_closure(self, list_of_candidate_sentences):
        if len(list_of_candidate_sentences) != 2:
            return False
        if abs(list_of_candidate_sentences[0].probability - list_of_candidate_sentences[1].probability) > self._tolerance:
            return False
        one_if_there_is_a_single_positive_closure = -1
        for a_statement in list_of_candidate_sentences:
            if POSITIVE_CLOSURE == a_statement.cardinality:
                one_if_there_is_a_single_positive_closure *= -1
        if one_if_there_is_a_single_positive_closure == 1:
            return True
        return False

    def _statement_for_a_group_with_a_useless_positive_closure(self, group_of_candidate_statements):
        for a_statement in group_of_candidate_statements:
            if a_statement.cardinality != POSITIVE_CLOSURE:
                return a_statement
        raise ValueError("The received group does not contain any statement with positive closure")

    def _group_IRI_constraints(self, candidate_statements):
        result = []
        already_visited = set()
        for i in range(0, len(candidate_statements)):
            a_statement = candidate_statements[i]
            if a_statement.st_property == self._instantiation_property_str:
                result.append(a_statement)
                already_visited.add(a_statement)
            else:  # a_statement.st_property != self._instantiation_property_str:
                if a_statement not in already_visited:
                    already_visited.add(a_statement)
                    group_to_decide = [a_statement]

                    for j in range(i + 1, len(candidate_statements)):
                        if self._statements_have_same_prop(a_statement,
                                                           candidate_statements[j]):
                            group_to_decide.append(candidate_statements[j])
                            already_visited.add(candidate_statements[j])
                    if len(group_to_decide) == 1:
                        result.append(a_statement)
                    else:  # At this point, group_to_decide may contain a list of constraints with the same property and
                           # different node constraint
                        if self._disable_or_statements:
                            for a_sentence in self._manage_group_to_decide_without_or(group_to_decide):
                                result.append(a_sentence)
                        else:
                            for a_sentence in self._manage_group_to_decide_with_or(group_to_decide):
                                result.append(a_sentence)

        return result

    def _statements_have_same_prop(self, st1, st2):
        if st1.st_property == st2.st_property:
            return True
        return False

    def _manage_group_to_decide_without_or(self, group_to_decide):
        """
        At this point, the candidate sentences can sahere prop, but no obj.
        This is, every sentence in group_to_decide has an unique obj.

        if len(group_to_decide) > 2 --> IRI should be picked, everything else to comments.
        if len(group_to_decide) == 2 --> IRI if it has higher trustworthiness, the specific obj otherwhise

        :param group_to_decide:
        :return:
        """
        result = []
        to_compose = []
        for a_statement in group_to_decide:
            if self._is_an_IRI(a_statement.st_type):
                to_compose.append(a_statement)
            else:
                result.append(a_statement)
        to_compose.sort(reverse=True, key=lambda x: x.probability)
        target_sentence = self._get_IRI_statement_in_group(to_compose)  # May be None
        self._remove_IRI_statements_if_useles(group_of_statements=to_compose)
        if len(to_compose) > 1:
            for a_statement in to_compose:
                if a_statement.st_type != IRI_ELEM_TYPE:
                    target_sentence.add_comment(self._turn_statement_into_comment(a_statement))
            result.append(target_sentence)
        elif len(to_compose) == 1:
            result.append(to_compose[0])
        # else  # No sentences to join

        return result

    def _manage_group_to_decide_with_or(self, group_to_decide):
        if not self._group_contains_IRI_statements(group_to_decide):
            for a_statement in group_to_decide:
                yield a_statement
        else:
            for a_new_statement in self._compose_statements_with_IRI_objects(group_to_decide):
                yield a_new_statement

    def _is_an_IRI(self, statement_type):
        return statement_type == IRI_ELEM_TYPE or statement_type.startswith("@")  # TODO careful here. Refactor


    def _remove_IRI_statements_if_useles(self, group_of_statements):
        # I am assuming a group of statements sorted by probability as param
        if len(group_of_statements) <= 1:
            return False
        index_of_IRI_statement = -1
        for i in range(0, len(group_of_statements)):
            if group_of_statements[i].st_type == IRI_ELEM_TYPE:
                index_of_IRI_statement = i
                break
        if index_of_IRI_statement != -1:
            if group_of_statements[1].probability == group_of_statements[index_of_IRI_statement].probability:
                # the previous 'if' works, trust me, im an engineer
                del group_of_statements[index_of_IRI_statement]
                return True
        return False

    def _group_contains_IRI_statements(self, list_of_candidate_statements):
        for a_statement in list_of_candidate_statements:
            if a_statement.st_type == IRI_ELEM_TYPE:
                return True
        return False

    def _compose_statements_with_IRI_objects(self, list_of_candidate_statements):
        result = []
        to_compose = []
        for a_statement in list_of_candidate_statements:
            if self._is_an_IRI(a_statement.st_type):
                to_compose.append(a_statement)
            else:
                result.append(a_statement)
        to_compose.sort(reverse=True, key=lambda x: x.probability)
        # target_probability = self._get_probability_of_IRI_statement_in_group(to_compose)
        iri_statement = self._get_IRI_statement_in_group(to_compose)
        was_removed_IRI = self._remove_IRI_statements_if_useles(to_compose)
        if not was_removed_IRI and not self._allow_redundant_or:  # The IRI macro is still there
            return [a_sentence for a_sentence in self._manage_group_to_decide_without_or(to_compose)] + result
        elif len(to_compose) > 1:  # There are some sentences to join in an OR and no IRI macro
            composed_statement = FixedPropChoiceStatement(
                st_property=to_compose[0].st_property,
                st_types=[a_statement.st_type for a_statement in to_compose],
                cardinality=POSITIVE_CLOSURE,
                probability=iri_statement.probability,
                n_occurences=iri_statement.n_occurences,
                serializer_object=self._statement_serializer_factory.get_choice_serializer(
                    is_inverse=to_compose[0].is_inverse
                ),
                is_inverse=to_compose[0].is_inverse
            )
            for a_statement in to_compose:
                if a_statement.st_type != IRI_ELEM_TYPE:
                    composed_statement.add_comment(self._turn_statement_into_comment(a_statement))
            result.append(composed_statement)
        elif len(to_compose) == 1:  # There is just one sentence in the group to join with OR
            result.append(to_compose[0])
        # else  # No sentences to join
        return result

    # def _get_probability_of_IRI_statement_in_group(self, group_of_statements):
    #     for a_statement in group_of_statements:
    #         if a_statement.st_type == IRI_ELEM_TYPE:
    #             return a_statement.probability
    #     raise ValueError("There is no IRI statement within the received group")

    def _get_IRI_statement_in_group(self, group_of_statements):
        for a_statement in group_of_statements:
            if a_statement.st_type == IRI_ELEM_TYPE:
                return a_statement
        return None

    def _statements_without_shapes_to_remove(self, original_statements, shape_names_to_remove):
        new_statements = []
        for a_statement in original_statements:
            if not a_statement.st_type in shape_names_to_remove:
                new_statements.append(a_statement)
        return new_statements



from plantuml import PlantUML
from shexer.utils.shapes import prefixize_shape_name_if_possible
from shexer.utils.uri import prefixize_uri_if_possible
from shexer.model.fixed_prop_choice_statement import FixedPropChoiceStatement
from shexer.consts import RDF_TYPE
import warnings


class UMLSerializer(object):

    def __init__(self, shapes_list, url_server, image_path, namespaces_dict=None, instantiation_property=RDF_TYPE):
        self._disable_connection_warnings()
        self._shapes_list = shapes_list
        self._url_server = url_server
        self._image_path = image_path
        self._instantiation_property = instantiation_property
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}

        self._shape_alias = {}

        self._diagram = None

        self._server_connection = self._init_server_connection()

    def serialize_shapes(self):

        self._reset_diagram()

        self._init_diagram()
        self._fill_diagram_with_shapes()
        self._close_diagram()
        result = self._send_diagram_to_server()
        self._store_diagram(result)

    def _disable_connection_warnings(self):
        warnings.filterwarnings("ignore", category=ResourceWarning)

    def _send_diagram_to_server(self):
        return self._server_connection.processes(self._diagram)


    def _store_diagram(self, img_diagram):
        with open(self._image_path, "wb") as out_stream:
            out_stream.write(img_diagram)


    def _fill_diagram_with_shapes(self):
        self._declare_shapes_and_atts()
        self._declare_relations()

    def _declare_relations(self):
        for a_shape in self._shapes_list:
            for a_statement in a_shape.yield_statements():
                if self._is_a_shape_link(a_statement):
                    origin_name = prefixize_shape_name_if_possible(a_shape_name=a_shape.name,
                                                                   namespaces_prefix_dict=self._namespaces_dict)
                    target_name = prefixize_shape_name_if_possible(a_shape_name=a_statement._st_type,
                                                                   namespaces_prefix_dict=self._namespaces_dict)
                    target_st = prefixize_uri_if_possible(target_uri=a_statement.st_property,
                                                          namespaces_prefix_dict=self._namespaces_dict,
                                                          corners=False)
                    self._write_line(
                        f"{self._shape_alias[origin_name]} --> {self._shape_alias[target_name]} : {target_st}")

    def _declare_shapes_and_atts(self):
        for a_shape in self._shapes_list:
            self._declare_and_open_shape(a_shape)
            self._declare_shape_atts(a_shape)
            self._close_shape()

    def _declare_shape_atts(self, a_shape):
        for a_statement in a_shape.yield_statements():
            if not self._is_a_shape_link(a_statement):
                prop = prefixize_uri_if_possible(target_uri=a_statement.st_property,
                                                 namespaces_prefix_dict=self._namespaces_dict,
                                                 corners=False)
                target_obj = self._serialize_obj_of_non_shape_link(a_statement)

                self._write_line(f"{prop} : {target_obj}"
                                 + f" {str(a_statement.cardinality) if a_statement.cardinality != 1 else ''}")


    def _serialize_obj_of_non_shape_link(self, a_statement):
        if not type(a_statement) == FixedPropChoiceStatement:
            result = prefixize_uri_if_possible(target_uri=a_statement.st_type,
                                         namespaces_prefix_dict=self._namespaces_dict,
                                         corners=False)
            if self._is_a_type_declaration(a_statement):
                result = "[" + result + "]"
            return result
        types = []
        for a_type in a_statement.st_types:
            if a_type.startswith("@"):
                types.append("@" + prefixize_uri_if_possible(target_uri=a_type[1:],
                                                             namespaces_prefix_dict=self._namespaces_dict,
                                                             corners=True))
            else:
                types.append(a_type)  # It should be IRI
        return " OR ".join(types)

    def _is_a_type_declaration(self, a_statement):
        return a_statement.st_property == self._instantiation_property

    def _is_a_shape_link(self, statement):
        if type(statement) == FixedPropChoiceStatement:
            return False
        return statement.st_type.startswith("@")

    def _declare_and_open_shape(self, a_shape):
        target_name = prefixize_shape_name_if_possible(a_shape_name=a_shape.name,
                                                       namespaces_prefix_dict=self._namespaces_dict)
        if ":" in target_name or "-" in target_name or target_name.startswith("<") :
            self._declare_shape_with_alias(target_name)
        else:
            self._declare_shape_without_alias(target_name)

    def _declare_shape_without_alias(self, target_name):
        self._shape_alias[target_name] = target_name
        self._write_line(f"object {target_name} {{")

    def _declare_shape_with_alias(self, target_name):
        alias = target_name
        alias = alias.replace(":", "_")
        alias = alias.replace("-", "_")

        if alias.startswith("<"):
            alias = target_name[1:-1]
        self._shape_alias[target_name] = alias
        self._write_line(f'object "{target_name}" as {alias} {{')

    def _close_shape(self):
        self._write_line("}", spacing=True)

    def _reset_diagram(self):
        self._diagram = ""

    def _init_diagram(self):
        self._write_line("@startuml", spacing=True)

    def _close_diagram(self):
        self._write_line("@enduml")

    def _write_line(self, text, spacing=False):
        self._diagram += text + "\n" * (1 if not spacing else 2)

    def _init_server_connection(self):
        return PlantUML(url=self._url_server)





from shexer.core.shexing.strategy.minimal_iri_strategy.abstract_min_iri_strategy import AbstractMinIriStrategy


class IgnoreMinIriStrategy(AbstractMinIriStrategy):

    def __init__(self):
        pass

    def annotate_shape_iri(self, shape):
        """
        Just skip this, no need to do anything

        :param shape:
        :param class_key:
        :return:
        """
        pass



from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import BaseFrequencyStrategy
from shexer.io.shex.formater.statement_serializers.frequency_strategy.abs_freq_serializer import AbsFreqSerializer
from shexer.io.shex.formater.statement_serializers.frequency_strategy.ratio_freq_serializer import RatioFreqSerializer

class MixedFrequencyStrategy(BaseFrequencyStrategy):

    def __init__(self, decimals=-1):
        self._abs_strategy = AbsFreqSerializer()
        self._ratio_strategy = RatioFreqSerializer(decimals=decimals)

    def serialize_frequency(self, statement):
        # The abs_strategy return a trailing dot that we want to skip. That why I use slicing here
        return self._ratio_strategy.serialize_frequency(statement) + \
               " (" + self._abs_strategy.serialize_frequency(statement)[:-1] + ")."





import unittest
from shexer.shaper import Shaper
from test.const import default_namespaces
from test.t_utils import number_of_shapes


class TestUrlEndpoint(unittest.TestCase):

    """
    These test cannot be too precise, since they are made against "live" content in Wikidata and DBpedia.

    """

    def test_wikidata(self):
        shape_map_raw = "SPARQL'select ?p where " \
                        "{ ?p <http://www.wikidata.org/prop/direct/P31> <http://www.wikidata.org/entity/Q14660> } " \
                        "LIMIT 1'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://query.wikidata.org/sparql",
                        namespaces_dict=default_namespaces(),
                        instantiation_property="http://www.wikidata.org/prop/direct/P31",
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 1)

    def test_dbpedia(self):
        shape_map_raw = "SPARQL'select ?s where {?s a <http://dbpedia.org/ontology/Person>} LIMIT 1'@<Flag>"
        shaper = Shaper(shape_map_raw=shape_map_raw,
                        url_endpoint="https://dbpedia.org/sparql",
                        namespaces_dict=default_namespaces(),
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        all_classes_mode=False)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) == 1)

    def test_all_classes_mode(self):
        # shape_map_raw = "SPARQL'select ?s where {?s a <http://dbpedia.org/ontology/Person>} LIMIT 1'@<Flag>"
        shaper = Shaper(all_classes_mode=True,
                        url_endpoint="https://agrovoc.fao.org/sparql",
                        namespaces_dict=default_namespaces(),
                        disable_comments=True,
                        depth_for_building_subgraph=1,
                        track_classes_for_entities_at_last_depth_level=False,
                        limit_remote_instances=5)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(number_of_shapes(str_result) > 2)
        pass  #



from shexer.model.shape_map import ShapeMap, ShapeMapItem
from shexer.io.shape_map.node_selector.node_selector_parser import NodeSelectorParser


class ListOfClassesToShapeMap(object):

    def __init__(self, sgraph, prefix_namespaces_dict):
        self._sgraph = sgraph
        self._selector_parser = NodeSelectorParser(prefix_namespaces_dict=prefix_namespaces_dict,
                                                   sgraph=sgraph)

    def str_class_list_to_shape_map_sparql_selectors(self, str_list, instantiation_property, limit_remote_instances):
        result = ShapeMap()
        instantiation_property = str(instantiation_property)
        for str_class in str_list:
            raw_selector = self._get_raw_selector_to_catch_instances_of_class_uri(class_uri=str_class,
                                                                                  instantiation_property=instantiation_property,
                                                                                  limit_remote_instances=limit_remote_instances)
            result.add_item(ShapeMapItem(node_selector=self._get_node_selector_object_for_raw_selector(raw_selector),
                                         shape_label=self._get_shape_label_for_class_uri(str_class)))
        return result

    def model_class_list_to_shape_map_sparql_selectors(self, obj_list, instantiation_property, limit_remote_instances):
        return self.str_class_list_to_shape_map_sparql_selectors(str_list=[str(an_elem) for an_elem in obj_list],
                                                                 instantiation_property=instantiation_property,
                                                                 limit_remote_instances=limit_remote_instances)

    def _get_shape_label_for_class_uri(self, class_uri):
        if "#" in class_uri and class_uri[-1] != "#":
            return class_uri[class_uri.rfind("#") + 1:]
        if "/" in class_uri:
            if class_uri[-1] != "/":
                return class_uri[class_uri.rfind("/") + 1:]
            else:
                return class_uri[class_uri[:-1].rfind("/") + 1:]
        else:
            return class_uri

    def _get_raw_selector_to_catch_instances_of_class_uri(self, class_uri, instantiation_property, limit_remote_instances):
        return 'SPARQL "select ?s where {{ ?s <{prop}> <{class_uri}> }} {limit}"'.format(
            class_uri=class_uri,
            prop=instantiation_property,
            limit="" if limit_remote_instances < 0 else "LIMIT " + str(limit_remote_instances)
        )
        # return '{' + 'FOCUS <{prop}> <{class_uri}>'.format(class_uri=class_uri, prop=instantiation_property) + '}'

    def _get_node_selector_object_for_raw_selector(self, raw_selector):
        return self._selector_parser.parse_node_selector(raw_selector=raw_selector)





STARTING_CHAR_FOR_SHAPE_NAME = "@"


class Shape(object):

    def __init__(self, name, class_uri, statements, n_instances):
        self._name = name
        self._class_uri = class_uri
        self._statements = statements if statements is not None else []
        self._n_instances = n_instances
        # self._inverse_statements = inverse_statements if inverse_statements is not None else []
        self._sorting_callback = lambda x: x.probability
        self._n_direct_statements = self._count_direct_statements(statements)
        self._n_inverse_statements = len(statements) - self._n_direct_statements

    @property
    def name(self):
        return self._name

    @property
    def class_uri(self):
        return self._class_uri

    @property
    def n_statements(self):
        return len(self._statements)

    @property
    def n_direct_statements(self):
        return self._n_direct_statements

    @property
    def n_inverse_statements(self):
        return self._n_inverse_statements

    @property
    def n_instances(self):
        return self._n_instances

    # @property
    # def iri_pattern(self):
    #     return self._iri_pattern
    #
    # @iri_pattern.setter
    # def iri_pattern(self, value):
    #     self._iri_pattern = value

    @property
    def statements(self):
        return self._statements

    @property
    def direct_statements(self):
        return [a_statement for a_statement in self._statements if not a_statement.is_inverse]

    @property
    def inverse_statements(self):
        return [a_statement for a_statement in self._statements if a_statement.is_inverse]

    @statements.setter
    def statements(self, value):
        self._statements = value

    @direct_statements.setter
    def direct_statements(self, statements):
        self._statements = [a_statement for a_statement in self._statements if a_statement.is_inverse]
        for a_statement in statements:
            self._statements.append(a_statement)

    @inverse_statements.setter
    def inverse_statements(self, inverse_statements):
        self._statements = [a_statement for a_statement in self._statements if not a_statement.is_inverse]
        for a_statement in inverse_statements:
            self._statements.append(a_statement)

    def yield_direct_statements(self):
        for a_statement in self._statements:
            if not a_statement.is_inverse:
                yield a_statement

    def yield_inverse_statements(self):
        for a_statement in self._statements:
            if a_statement.is_inverse:
                yield a_statement

    def sort_statements(self, callback, reverse=False):
        self._statements.sort(key=lambda x: callback(x), reverse=reverse)


    def yield_statements(self, just_direct=False):
        if just_direct:
            for a_statement in self.yield_direct_statements():
                yield a_statement
        else:
            for a_statement in self._statements:
                yield a_statement

    @staticmethod
    def _count_direct_statements(statements):
        counter = 0
        for a_statement in statements:
            if not a_statement.is_inverse:
                counter += 1
        return counter




import json


class AbstractProfileSerializer(object):

    def __init__(self, profile_obj):
        self._profile_obj = profile_obj

    def write_profile_to_file(self, target_file):
        with open(target_file, "w") as out_stream:
            json.dump(self._profile_obj, out_stream, indent=2)

    def get_string_representation(self):
        return json.dumps(self._profile_obj, indent=2)



from rdflib import URIRef, Graph, Literal, BNode
from shexer.model.graph.abstract_sgraph import SGraph
from shexer.utils.triple_yielders import tune_token, tune_prop, tune_subj
from shexer.utils.uri import add_corners_if_it_is_an_uri, remove_corners
from shexer.core.instances.pconsts import _S, _P, _O
from shexer.model.IRI import IRI as ModelIRI
from shexer.model.property import Property as ModelProperty
from shexer.model.Literal import Literal as ModelLiteral
from shexer.model.bnode import BNode as ModelBnode
from shexer.consts import RDF_TYPE



class RdflibSgraph(SGraph):

    def __init__(self, rdflib_graph=None, source_file=None, raw_graph=None, format="turtle"):
        """
        Pass an rdflib.Graph object or the params source_file and format to parse a local rdf

        :param rdflib_graph:
        :param source_file:
        :param format:
        """
        super().__init__()
        self._rdflib_graph = rdflib_graph if rdflib_graph is not None else self._build_rdflib_graph(source=source_file,
                                                                                                    raw_graph=raw_graph,
                                                                                                    format=format)

    def query_single_variable(self, str_query, variable_id):
        rows_res = self._rdflib_graph.query(str_query)
        return [str(a_row[0]) for a_row in rows_res]

    def serialize(self, path, format):
        self._rdflib_graph.serialize(destination=path,
                                     format=format)

    def yield_p_o_triples_of_an_s(self, target_node):
        for s, p ,o in self._rdflib_graph.triples((URIRef(remove_corners(a_uri=target_node,
                                                                         raise_error_if_no_corners=False)),
                                                   None,
                                                   None)):
            yield self._add_URI_corners_if_needed(s),\
                  self._add_URI_corners_if_needed(p),\
                  self._add_URI_corners_if_needed(self._add_lang_if_needed(o))

    def yield_s_p_triples_of_an_o(self, target_node):
        for s, p, o in self._rdflib_graph.triples((None,
                                                   None,
                                                   URIRef(remove_corners(a_uri=target_node,
                                                                         raise_error_if_no_corners=False)))):
            yield self._add_URI_corners_if_needed(s),\
                  self._add_URI_corners_if_needed(p),\
                  self._add_URI_corners_if_needed(o)


    def yield_class_triples_of_an_s(self, target_node, instantiation_property):
        for s ,p, o in self._rdflib_graph.triples((URIRef(remove_corners(a_uri=target_node,
                                                                         raise_error_if_no_corners=False)),
                                                   URIRef(remove_corners(a_uri=instantiation_property,
                                                                         raise_error_if_no_corners=False)),
                                                   None)):
            yield self._add_URI_corners_if_needed(s),\
                  self._add_URI_corners_if_needed(p),\
                  self._add_URI_corners_if_needed(self._add_lang_if_needed(o))

    def add_triple(self, a_triple):
        """
        It receives a tuple of 3 string elements. It adds it to the local rdflib graph
        :param a_triple:
        :return:
        """

        subj = tune_subj(add_corners_if_it_is_an_uri(a_triple[_S]),
                         raise_error_if_no_corners=False)
        prop = tune_prop(add_corners_if_it_is_an_uri(a_triple[_P]),
                         raise_error_if_no_corners=False)
        obj = tune_token(add_corners_if_it_is_an_uri(a_triple[_O]),
                         raise_error_if_no_corners=False)

        self._rdflib_graph.add((self._turn_obj_into_rdflib_element(subj),
                                self._turn_obj_into_rdflib_element(prop),
                                self._turn_obj_into_rdflib_element(obj)))

    def yield_classes_with_instances(self, instantiation_property=RDF_TYPE):
        result = set()
        for s ,_, _ in self._rdflib_graph.triples((None,
                                                   URIRef(remove_corners(a_uri=instantiation_property,
                                                                         raise_error_if_no_corners=False)),
                                                   None)):
            result.add(str(s))
        for elem in result:
            yield elem


    def _turn_obj_into_rdflib_element(self, model_elem):
        if type(model_elem) == ModelIRI or type(model_elem) == ModelProperty:
            return URIRef(model_elem.iri)
        elif type(model_elem) == ModelLiteral:
            return Literal(lexical_or_value=str(model_elem),
                           datatype=model_elem.elem_type)
        elif type(model_elem) == ModelBnode:
            return BNode(value=str(model_elem))
        else:
            raise ValueError("Unexpected type of element. " + str(model_elem) + ": " + str(type(model_elem)))


    def _build_rdflib_graph(self, source, raw_graph, format):
        result = Graph()
        if source is not None:
            result.parse(source=source, format=format)
        else:
            result.parse(data=raw_graph, format=format)
        return result

    def _add_lang_if_needed(self, rdflib_obj):
        """
        It returns a string representation with lang if it is a langString
        :param rdflib_obj:
        :return:
        """
        if type(rdflib_obj) == Literal and rdflib_obj.language is not None:
           return '"' + str(rdflib_obj) + '"@' + rdflib_obj.language
        return rdflib_obj

    def _add_URI_corners_if_needed(self, rdflib_obj):
        if type(rdflib_obj) == URIRef:
            return "<"+ str(rdflib_obj) + ">"
        return str(rdflib_obj)


from shexer.model.const_elem_types import BNODE_ELEM_TYPE

class BNode(object):

    def __init__(self, identifier):
        self._identifier = identifier

    def __str__(self):
        return self._identifier

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        return str(self) == str(other)

    @property
    def elem_type(self):
        return BNODE_ELEM_TYPE

    @property
    def iri(self):
        return self._identifier



from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME
from shexer.utils.uri import prefixize_uri_if_possible
from shexer.io.shex.formater.consts import SHAPE_LINK_CHAR

def build_shapes_name_for_class_uri(class_uri, shapes_namespace):

    if class_uri.startswith("@"): # special shape case
        return class_uri
    if class_uri.startswith("<") and class_uri.endswith(">"):
        return STARTING_CHAR_FOR_SHAPE_NAME + class_uri
    last_piece = class_uri
    if "#" in last_piece and last_piece[-1] != "#":
        last_piece = last_piece[last_piece.rfind("#") + 1:]
    if "/" in last_piece:
        if last_piece[-1] != "/":
            last_piece = last_piece[last_piece.rfind("/") + 1:]
        else:
            last_piece = last_piece[last_piece[:-1].rfind("/") + 1:]
    if last_piece.endswith(">"):
        last_piece = last_piece[:-1]
    if last_piece.startswith("<"):
        last_piece = last_piece[1:]
    return STARTING_CHAR_FOR_SHAPE_NAME + "<" + shapes_namespace + last_piece + ">" if last_piece is not None else class_uri
        # return class_uri


def build_shape_name_for_qualifier_prop_uri(prop_uri, shapes_namespace):  # TODO REVIEW!
    last_piece = None
    if "#" in prop_uri and prop_uri[-1] != "#":
        last_piece = prop_uri[prop_uri.rfind("#") + 1:]
    if "/" in prop_uri:
        if prop_uri[-1] != "/":
            last_piece = prop_uri[prop_uri.rfind("/") + 1:]
        else:
            last_piece = prop_uri[prop_uri[:-1].rfind("/") + 1:]
    if last_piece is not None:
        return STARTING_CHAR_FOR_SHAPE_NAME + "<" + shapes_namespace + last_piece + ">"
    return STARTING_CHAR_FOR_SHAPE_NAME + prop_uri.upper()


def prefixize_shape_name_if_possible(a_shape_name, namespaces_prefix_dict):
    result = prefixize_uri_if_possible(target_uri=a_shape_name[1:],                  # Avoid the "@" starting char
                                       namespaces_prefix_dict=namespaces_prefix_dict)
    return result


from shexer.core.instances.pconsts import _S, _P, _O
from shexer.core.instances.annotators.strategy_mode.target_classes_mode import TargetClassesMode
from shexer.core.instances.annotators.strategy_mode.all_classes_mode import AllClasesMode
from shexer.core.instances.annotators.strategy_mode.shape_qualifiers_mode import ShapeQualifiersMode
from shexer.core.instances.annotators.strategy_mode.compound_strategy_mode import CompoundStrategyMode
from shexer.core.instances.annotators.strategy_mode.instance_cap_mode import InstanceCapMode


class BaseAnnotator(object):

    def __init__(self, instance_tracker):
        self._instance_tracker = instance_tracker

        # Some short-path to avoid verbosity and too deep references. We'll be gentle with private stuff ;)
        self._all_classes_mode = self._instance_tracker._all_classes_mode
        self._instantiation_property = self._instance_tracker._instantiation_property
        self._subclass_property = self._instance_tracker._subclass_property
        self._instances_dict = self._instance_tracker._instances_dict
        self._classes_considered_in_htree = self._instance_tracker._classes_considered_in_htree
        self._htree = self._instance_tracker._htree
        self._shape_qualifiers_mode = self._instance_tracker._shape_qualifiers_mode
        self._namespaces_for_qualifiers_props = self._instance_tracker._namespaces_for_qualifiers_props
        self._target_classes = self._instance_tracker._target_classes
        self._shapes_namespace = self._instance_tracker._shapes_namespace
        self._instances_cap = self._instance_tracker._instances_cap

        self._strategy_mode = self._get_proper_strategy()

    def is_relevant_triple(self, a_triple):
        return self._strategy_mode.is_relevant_triple(a_triple)

    def annotate_triple(self, a_triple):
        self._strategy_mode.annotate_triple(a_triple)

    def add_instance_to_instances_dict(self, a_triple):
        if a_triple[_S].iri not in self._instances_dict:
            self._instances_dict[a_triple[_S].iri] = []

    # def annotate_class(self, a_triple):
    #     self._strategy_mode.annotate_class(a_triple)

    def annotation_post_parsing(self):
        self._strategy_mode.annotation_post_parsing()


    def _get_proper_strategy(self):
        result = None
        strategies_list = []
        target_classes_flag = False
        if self._all_classes_mode:
            strategies_list.append(AllClasesMode(annotator_ref=self))
        if self._target_classes is not None and len(self._target_classes) > 0:
            strategies_list.append(TargetClassesMode(annotator_ref=self))
            target_classes_flag = True
        if self._shape_qualifiers_mode:
            strategies_list.append(
                ShapeQualifiersMode(annotator_ref=self,
                                    namespaces_for_qualifiers_props=self._namespaces_for_qualifiers_props,
                                    shapes_namespace=self._shapes_namespace))

        if len(strategies_list) == 0:
            raise ValueError("Wrong combination of params when building the instance tracker. There are not target classes")
        if len(strategies_list) == 1:
            result = strategies_list[0]
        else:
            result = CompoundStrategyMode(annotator_ref=self,
                                        list_of_strategies=strategies_list)
        return result if self._instances_cap <= 0 else \
            InstanceCapMode(annotator_ref=self,
                            internal_strategy=result,
                            instance_limit=self._instances_cap,
                            n_target_classes=-1 if not target_classes_flag or len(strategies_list) > 1
                                                   else len(self._target_classes)
                            )



from shexer.model.statement import Statement

class FixedPropChoiceStatement(Statement):

    def __init__(self, st_property, st_types, cardinality, n_occurences, probability, comments=None,
                 serializer_object=None, is_inverse=False):
        super(FixedPropChoiceStatement, self).__init__(st_property=st_property,
                                                       st_type=None,
                                                       cardinality=cardinality,
                                                       n_occurences=n_occurences,
                                                       probability=probability,
                                                       comments=comments,
                                                       serializer_object=serializer_object,
                                                       is_inverse=is_inverse)
        self._st_types = st_types

    @property
    def st_type(self):
        raise TypeError("Choice statements doesnt have a single type")

    @property
    def st_types(self):
        return self._st_types



_TRACKERS_DISAM_COUNT = 0

class AbstractInstanceTracker(object):

    def track_instances(self, verbose=False):
        raise NotImplementedError()


    @property
    def disambiguator_prefix(self):
        """
        It return an str that may help for disambiguation purposes if the instance_tracker is used to produce dicts
        that may be integrated with other instance dicts and there should be any key colission.
        :return:
        """
        global _TRACKERS_DISAM_COUNT
        _TRACKERS_DISAM_COUNT += 1
        return self._specific_disambiguator_prefix() + str(_TRACKERS_DISAM_COUNT )

    def _specific_disambiguator_prefix(self):
        raise NotImplementedError()


from shexer.model.const_elem_types import IRI_ELEM_TYPE, LITERAL_ELEM_TYPE, BNODE_ELEM_TYPE

class HTree(object):
    def __init__(self):
        self._root = None  # HNode
        self._node_index = {}

    def _subscribe_element(self, hcontent):
        str_value = hcontent.str_value
        if str_value not in self._node_index:
            self._node_index[str_value] = hcontent

    def create_node_literal(self, literal_obj):
        return HNode(hcontent=HLiteral(value=literal_obj),
                     htree=self)

    def create_node_IRI(self, iri_obj):
        return HNode(hcontent=HIri(value=iri_obj),
                     htree=self)

    def create_node_macro(self, macro_obj):
        return HNode(hcontent=HMacro(value=macro_obj),
                     htree=self)

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, value):
        self._root = value

    @property
    def iri_node(self):
        return None if not self.contains_element(IRI_ELEM_TYPE) else self.get_node_of_element(IRI_ELEM_TYPE)

    @property
    def literal_node(self):
        return None if not self.contains_element(LITERAL_ELEM_TYPE) else self.get_node_of_element(LITERAL_ELEM_TYPE)

    @property
    def bnode_node(self):
        return None if not self.contains_element(BNODE_ELEM_TYPE) else self.get_node_of_element(BNODE_ELEM_TYPE)

    def contains_element(self, str_type):
        return str_type in self._node_index

    def get_node_of_element(self, str_type):
        return self._node_index[str_type]  # We could get a key violation error here. I decide to risk that
        # assuming that whoever calls this has checked contains_element()



class HNode(object):
    def __init__(self, hcontent, htree, parents=None, children=None):
        self._hcontent = hcontent
        self._parents = parents if parents is not None else {}
        self._children = children if children is not None else {}
        htree._subscribe_element(self)  # Eeeasy python conventions! everything is under control

    def add_child(self, child):
        str_child = child.str_value
        if str_child not in self._children:
            self._children[str_child] = child
            child._parents[self.str_value] = self  # Lets assume consistence


    def add_parent(self, parent):
        str_parent = parent.str_value
        if str_parent not in self._parents:
            self._parents[str_parent] = parent
            parent._children[self.str_value] = self  # Lets assume consistence

    def has_parents(self):
        if self._parents:
            return True
        return False


    @property
    def value(self):
        return self._hcontent.value

    @property
    def str_value(self):
        return self._hcontent.str_value

    def __eq__(self, other):
        if not isinstance(other, HNode):
            return False
        return self._hcontent == other._hcontent

    def __ne__(self, other):
        return not self.__eq__(other)


class HContent(object):  # Abstract, dont instantiate
    def __init__(self, value):
        self._value = value

    @property
    def str_value(self):
        return str(self._value)

    @property
    def value(self):
        return self._value

    def __eq__(self, other):
        if not isinstance(other, HContent):
            return False
        if type(self._value) != type(other._value):
            return False
        return self.str_value == other.str_value

    def __ne__(self, other):
        return not self.__eq__(other)


class HLiteral(HContent):
    def __init__(self, value):  # Value should be a Literal
        super(HLiteral, self).__init__(value)

    def __str__(self):
        return str(self.value)


class HMacro(HContent):
    def __init__(self, value):
        super(HMacro, self).__init__(value)

    def __str__(self):
        return str(self.value)


class HIri(HContent):
    def __init__(self, value):
        super(HIri, self).__init__(value)

    def __str__(self):
        return str(self.value)






from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder


class MultiZipTriplesYielder(BaseTriplesYielder):

    def __init__(self, multiyielders):
        super().__init__()
        self._multiyielders = multiyielders

        self._triples_yielded_from_used_yielders = 0
        self._error_triples_from_used_yielders = 0
        self._last_yielder = None

        self._current_yielder = None

    def yield_triples(self, parse_namespaces=True):
        self._reset_count()
        for a_yielder in self._multiyielders:
            self._current_yielder = a_yielder
            for a_triple in a_yielder.yield_triples():
                yield a_triple
            self._triples_yielded_from_used_yielders += a_yielder.yielded_triples
            self._error_triples_from_used_yielders += a_yielder.error_triples

    @property
    def yielded_triples(self):
        triples_current_yielder = 0 if self._current_yielder is None else self._current_yielder.yielded_triples
        return self._triples_yielded_from_used_yielders + triples_current_yielder

    @property
    def error_triples(self):
        errors_current_yielder = 0 if self._current_yielder is None else self._current_yielder.error_triples
        return self._error_triples_from_used_yielders + errors_current_yielder

    def _reset_count(self):
        self._error_triples_from_used_yielders = 0
        self._triples_yielded_from_used_yielders = 0




from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import BaseFrequencyStrategy

class AbsFreqSerializer(BaseFrequencyStrategy):

    def serialize_frequency(self, statement):
        return str(statement.n_occurences) + " instance{}.".format(
            "" if statement.n_occurences == 1
            else "s"
        )








XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema#"
XSD_PREFIX = "xsd"

RDF_SYNTAX_NAMESPACE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDF_PREFIX = "rdf"
RDF_TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

DT_NAMESPACE = "http://dbpedia.org/datatype/"
DT_PREFIX = "dt"

OPENGIS_NAMESPACE = "http://www.opengis.net/ont/geosparql#"
OPENGIS_PREFIX = "geo"

LANG_STRING_TYPE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
STRING_TYPE = "http://www.w3.org/2001/XMLSchema#string"
FLOAT_TYPE = "http://www.w3.org/2001/XMLSchema#float"
INTEGER_TYPE = "http://www.w3.org/2001/XMLSchema#integer"


def _add_prefix(unprefixed_elem, prefix):
    return prefix + ":" + unprefixed_elem


def remove_corners(a_uri, raise_error_if_no_corners=True):
    if a_uri.startswith("<") and a_uri.endswith(">"):
        return a_uri[1:-1]
    if raise_error_if_no_corners:
        raise ValueError("Wrong parameter of function: '" + a_uri + "'")
    else:
        return a_uri


def add_corners(a_uri):
    return "<" + a_uri + ">"

def add_corners_if_needed(a_uri):
    if a_uri.startswith("<"):
        return a_uri
    return add_corners(a_uri)

def longest_common_prefix(uri1, uri2):
    """
    It returns an str containing the longest possible common initial part of uri1 and uri2

    :param uri1:
    :param uri2:

    :return:
    """

    if len(uri1) == 0 or len(uri2) == 0:
        return ""
    shortest = len(uri1) if len(uri1) < len(uri2) else len(uri2)
    for i in range(shortest):
        if uri1[i] != uri2[i]:
            return uri1[:i]
    return uri1[:shortest]

def add_corners_if_it_is_an_uri(a_candidate_uri):
    if a_candidate_uri.startswith("http://") or a_candidate_uri.startswith("https://"):  # TODO, check this!
        return "<" + a_candidate_uri + ">"
    return a_candidate_uri


def decide_literal_type(a_literal, base_namespace=None):
    if there_is_arroba_after_last_quotes(a_literal):
        return LANG_STRING_TYPE
    elif "\"^^" not in a_literal:
        return STRING_TYPE
    elif "xsd:" in a_literal:
        return XSD_NAMESPACE + a_literal[a_literal.find("xsd:") + 4:]
    elif "rdf:" in a_literal:
        return RDF_SYNTAX_NAMESPACE + a_literal[a_literal.find("rdf:")+ 4:]
    elif "dt:" in a_literal:
        return DT_NAMESPACE + a_literal[a_literal.find("dt:")+ 3:]
    elif "geo:" in a_literal:
        return OPENGIS_NAMESPACE + a_literal[a_literal.find("geo:") + 4:]
    elif XSD_NAMESPACE in a_literal or RDF_SYNTAX_NAMESPACE in a_literal \
            or DT_NAMESPACE in a_literal or OPENGIS_NAMESPACE in a_literal:
        return a_literal[a_literal.find("\"^^")+4:-1]
    elif a_literal.strip().endswith(">"):
        candidate_type = a_literal[a_literal.find("\"^^") + 4:-1]  # plain uri, no corners
        if base_namespace is not None and not candidate_type.startswith("http"):
            return base_namespace + candidate_type
        return candidate_type
    else:
        raise RuntimeError("Unrecognized literal type:" + a_literal)


def is_a_correct_uri(target_uri, prefix_namespace_dict):
    """
    TODO: Here I am assuming that there is no forbiden char ( " < > # % { } | \ ^ ~ [ ] ` )
    :param target_uri:
    :param prefix_namespace_dict:
    :return:
    """
    if target_uri[0] == "<" and target_uri[-1] == ">":
        return True
    for a_prefix in prefix_namespace_dict:
        if target_uri.startswith(a_prefix + ":"):
            return True
        return False


def there_is_arroba_after_last_quotes(target_str):
    if target_str.rfind("@") > target_str.rfind('"'):
        return True
    return False


def parse_literal(an_elem, base_namespace=None):
    content = an_elem[1:an_elem.find('"', 1)]
    elem_type = decide_literal_type(a_literal=an_elem,
                                    base_namespace=base_namespace)
    return content, elem_type

def parse_unquoted_literal(an_elem):
    elem_type = decide_literal_type(an_elem)
    return an_elem, elem_type


def unprefixize_uri_if_possible(target_uri, prefix_namespaces_dict, include_corners=True):
    for a_prefix in prefix_namespaces_dict:
        if target_uri.startswith(a_prefix+":"):
            result = target_uri.replace(a_prefix+":", prefix_namespaces_dict[a_prefix])
            if include_corners:
                result = add_corners(result)
            return result
    return target_uri

def unprefixize_uri_mandatory(target_uri, prefix_namespaces_dict, include_corners=True):
    for a_prefix in prefix_namespaces_dict:
        if target_uri.startswith(a_prefix+":"):
            result = target_uri.replace(a_prefix+":", prefix_namespaces_dict[a_prefix])
            if include_corners:
                result = add_corners(result)
            return result
    raise ValueError("Unrecognized prefix in the following element" + target_uri)


def prefixize_uri_if_possible(target_uri, namespaces_prefix_dict, corners=True):
    best_match = None
    candidate_uri = remove_corners(target_uri) if corners else target_uri
    for a_namespace in namespaces_prefix_dict:  # Prefixed element (all literals are prefixed elements)
        if candidate_uri.startswith(a_namespace):
            if "/" not in candidate_uri[len(a_namespace):] and \
                "#" not in candidate_uri[len(a_namespace):]:
                best_match = a_namespace
                break
            # if best_match is None or len(best_match) < len(a_namespace):
            #     best_match = a_namespace

    return target_uri if best_match is None else candidate_uri.replace(best_match, namespaces_prefix_dict[best_match] + ":")









import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE_ITER


_BASE_DIR = BASE_FILES + "instances_cap" + pth.sep


class TestInstancesCap(unittest.TestCase):
    """
    We must use a deterministic input format, such as NT or TURTLE_ITER (something non rdflib-based)
    to test this. Otherwhise, the instance(s) chosen for each test could be different in different
    executions of the tests

    """

    def test_no_cap_all_g1(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=-1)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_cap_1_all_g1(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=1)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "all_classes_cap_1.shex",
                                                      str_target=str_result))

    def test_cap_3_all_g1(self):  # exceeds one class limit (of instances), doest reach the other class limit
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=3)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "all_classes_cap_3.shex",
                                                      str_target=str_result))

    def test_cap_5_all_g1(self):  # exceeds every class limit (of instances), doest reach the other class limit
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=5)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_cap_1_one_target_g1(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=False,
            target_classes=["foaf:Person"],
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=1)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "target_person_cap_1.shex",
                                                      str_target=str_result))


    def test_cap_5_one_target_g1(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=False,
            target_classes=["foaf:Person"],
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=5)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "target_person_cap_5.shex",
                                                      str_target=str_result))

    def test_cap_1_all_targets_g1(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=False,
            target_classes=["foaf:Person", "foaf:Document"],
            input_format=TURTLE_ITER,
            disable_comments=True,
            instances_cap=1)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "all_classes_cap_1.shex",
                                                      str_target=str_result))






from shexer.io.shex.formater.statement_serializers.frequency_strategy.base_frequency_strategy import BaseFrequencyStrategy

class RatioFreqSerializer(BaseFrequencyStrategy):

    def __init__(self, decimals=-1):
        """

        :param decimals: it indicates the number of decimals to use to express ratios.
                        When a negative number is provided, decimals won't be controlled
        """
        self._decimals=decimals
        if decimals < 0:
            self.serialize_frequency = self._serialize_freq_unbounded
        elif decimals ==0:
            self.serialize_frequency = self._serialize_freq_int
        else:
            self.serialize_frequency = self._serialize_freq_decimals


    def serialize_frequency(self, statement):
        raise NotImplementedError("This function will be initialized with a callback during the __init__")

    def _serialize_freq_unbounded(self, statement):
        return str(statement.probability * 100) + " %"

    def _serialize_freq_decimals(self, statement):
        pattern = "{:." + str(self._decimals) +"f} %"
        return pattern.format(statement.probability*100)

    def _serialize_freq_int(self, statement):
        return str(int(statement.probability * 100)) + " %"



import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces
from test.t_utils import graph_comparison_file_vs_str
import os.path as pth
from shexer.consts import TURTLE, SHACL_TURTLE

_BASE_DIR = BASE_FILES + "shacl" + pth.sep


class TestClassSelection(unittest.TestCase):

    def test_all_classes_mode(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "g1_all_classes.ttl",
                                                     str_target=str_result))

    def test_all_classes_with_target_classes(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=False,
            target_classes=["http://xmlns.com/foaf/0.1/Person",
                            "http://xmlns.com/foaf/0.1/Document"],
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "g1_all_classes.ttl",
                                                     str_target=str_result))

    def test_some_classes_with_target_classes(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=False,
            target_classes=["http://xmlns.com/foaf/0.1/Person"],
            input_format=TURTLE,
            disable_comments=True)
        str_result = shaper.shex_graph(string_output=True,
                                       output_format=SHACL_TURTLE)
        self.assertTrue(graph_comparison_file_vs_str(file_path=_BASE_DIR + "g1_person.ttl",
                                                     str_target=str_result))



from distutils.core import setup
from setuptools import find_packages

def read(file_path):
	with open(file_path, "r") as in_stream:
		return in_stream.read()

setup(
  name = 'shexer',
  packages = find_packages(exclude=["*.local_code.*"]), # this must be the same as the name above
  version = '2.5.5',
  description = 'Automatic schema extraction for RDF graphs',
  author = 'Daniel Fernandez-Alvarez',
  author_email = 'danifdezalvarez@gmail.com',
  url = 'https://github.com/DaniFdezAlvarez/shexer',
  download_url = 'https://github.com/DaniFdezAlvarez/shexer/archive/2.5.5.tar.gz',
  keywords = ['testing', 'shexer', 'shexerp3', "rdf", "shex", "shacl", "schema"],
  long_description = read('README.md'),
  long_description_content_type='text/markdown',
  classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
		"Programming Language :: Python :: 3.11",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
  install_requires=[            
          'Flask',
          'Flask-Cors',
		  'rdflib',
		  'SPARQLWrapper',
          'wlighter',
          'plantuml',
          'python-xz'
      ],
)



from shexer.model.shape import STARTING_CHAR_FOR_SHAPE_NAME

class ShapeMapLabelParser(object):

    def __init__(self, prefix_namespaces_dict=None):
        self._namespaces_prefix_dict = prefix_namespaces_dict if prefix_namespaces_dict is not None else {}

    def parse_shape_map_label(self, raw_label):

        if self._is_a_prefixed_uri(raw_label):
            return STARTING_CHAR_FOR_SHAPE_NAME + self._parse_prefixed_label(raw_label)
        return STARTING_CHAR_FOR_SHAPE_NAME + raw_label
        # return self._parse_unprefixed_label(raw_label)


    def _is_a_prefixed_uri(self, raw_label):
        if len(raw_label) < 2:
            return False
        if raw_label.startswith("<") and raw_label.endswith(">"):
            return False
        return True


    # def _parse_unprefixed_label(self, raw_label):
    #     return raw_label[1:-1]

    def _parse_prefixed_label(self, raw_label):
        index_sep = raw_label.find(":")
        if index_sep == -1:
            raise ValueError("Wrong label: expecting a URI surrounded by <> or a prefixed element: " + raw_label)
        target_prefix = raw_label[:index_sep]
        if target_prefix in self._namespaces_prefix_dict:
            return self._namespaces_prefix_dict[target_prefix] + raw_label[index_sep + 1:]
        else:
            raise ValueError("Unknown prefix in label: " + raw_label)




